--[[
	Core:
	* Profile and default settings
	* Loading of bars
	* Unit backbone
]]
local debug = false

-- Libs
local AceDB = LibStub("AceDB-3.0")

local mHud = LibStub("AceAddon-3.0"):NewAddon("mHud")
local mBar = LibStub("AceAddon-3.0"):GetAddon("mBar")

BINDING_HEADER_MHUD = "mHud"
BINDING_NAME_TOGGLEMHUD = "Enable or disable mHud"

mHud.bars = { }
mHud.units = { }
mHud.onUpdateUnits = { }
mHud.events = {
	"UNIT_PET",
	"UNIT_TARGET",
	"PLAYER_FOCUS_CHANGED",
	"PLAYER_TARGET_CHANGED",
	"CHANNEL_ROSTER_UPDATE",
	"UPDATE_SHAPESHIFT_FORM",
	"PLAYER_ENTERING_WORLD",
	"PLAYER_REGEN_DISABLED",
	"PLAYER_REGEN_ENABLED",
}

mHud.defaults = {
	profile = {
		ver = 0.02,
		skin = nil,
		
		unitFPS = 10,
		
		module = { ['*'] = { active = true },},
		bar = {
			['*'] = {
				active = false,
				
				unit = "player",
				type = "Health",
				
				height = 250,
				width = 150,

				offsetX = 0,
				offsetY = 0,
				texture = "Blizzard",
				
				rotation = 3,
				textureRot = 0,
				textureFlip = false,
				textureMirror = false,
				
				frameLevel = 1,
				
				anchor = {
					to = "UIParent",
					point = "CENTER",
					relative = "CENTER",
				},
			},
		},
	},
}

function mHud:OnEnable()

	-- Frame for events and OnUpdate
	if not self.event then self.event = CreateFrame("Frame") end
	
	-- Load db
	if self.skin then self.skin:Init() end -- Adds to mHud.defaults
	if not self.db then
		self.db = AceDB:New("mHudDB", mHud.skin.defaults, "Default")
	else
		self.db:RegisterDefaults(mHud.skin.defaults)
	end
	if self.skin then self.skin:Load() end -- Load default profile if none is set
	
	-- Load modules
	for name, module in self:IterateModules() do
		if self.db.profile.module[name].active ~= false then self:EnableModule(name) end
	end
	
	-- Load bars
	for id, db in pairs(self.db.profile.bar) do
		if db.active then self.bars[id] = mBar:CreateBar(id, db) end
	end
	for id, f in pairs(self.bars) do mBar:Anchor(f) end -- Update all anchors
	
	self:RegisterUnits()
	self:RegisterEvents()
	
	-- Get check for existance for all units
	for unit, _ in pairs(self.units or { }) do
		self:UpdateBarsForUnit(unit)
	end
	
end


function mHud:OnDisable()

	-- Disable modules
	for name, module in self:IterateModules() do
		if self:IsEnabled(name) then self:DisableModule(name) end
	end
	
	-- Disable all bars
	for id, f in pairs(self.bars) do
		mBar:RemoveBar(f)
		self.bars[id] = nil
	end
	
	self:UnregisterEvents()
	
end


function mHud:RegisterEvents()

	self:Debug("RegisterEvents")
	if self.events then
		for _, e in pairs(self.events) do self.event:RegisterEvent(e) end
	end
	
	self.event:SetScript("OnEvent", function() self:OnEvent() end)
	
end

function mHud:UnregisterEvents()

	self:Debug("UnregisterEvents")
	if self.events then
		for _, e in pairs(self.events) do self.event:UnregisterEvent(e) end
	end
	
end


function mHud:RegisterUnits()
	
	self.units = { } -- Reset table, only done on config change / load, so don't care about the GC
	self.onUpdateUnits = { }
	
	for _, f in pairs(self.bars) do
		--UnregisterUnitWatch(f)
		--f:SetAttribute("unit", string.lower(f.db.unit))
		--f:SetScript("OnShow", function() self:Debug("SHOW", true); self:UpdateBarsForUnit(f.db.unit) end)
		--RegisterUnitWatch(f)

		local unit = string.lower(f.db.unit)
		
		if not self.units[unit] then self.units[unit] = { } end
		table.insert(self.units[unit], f)
				
		if unit ~= "target" and unit ~= "playertarget" then
		    local start = string.find(unit, 'target')
		    if unit == "targettarget" or (start and start > 2) then
				table.insert(self.onUpdateUnits, {unit = unit, unitName = "", health = "", power = ""}) -- Update to GUID 2.4
		    end
		end
		
	end
	
	if # self.onUpdateUnits > 0 then
		self.event.lastRun = 0
		self.event:SetScript("OnUpdate", function() self:OnUpdate() end)
	else
		self.event:SetScript("OnUpdate", nil)
	end
	
end


function mHud:UpdateBarsForUnit(unit)

	local visible = UnitExists(unit)
	for _, f in pairs(self.units[unit] or { }) do
		if visible and f.db.active == true then
			--self:Debug("Show: "..f.id, true)
			f:Show()
			mBar:SendMessage("MBAR_UPDATE", f)
		else
			--self:Debug("Hide: "..f.id, true)
			f:Hide()
		end
	end
	
end


function mHud:OnEvent()
	
	self:Debug("OnTargetEvent: "..event.." arg1: "..(arg1 or "nil"))
	if event == "PLAYER_TARGET_CHANGED" then
		self:UpdateBarsForUnit("target")
		self:UpdateBarsForUnit("targettarget")
		
	elseif event == "UPDATE_SHAPESHIFT_FORM" then
		self:UpdateBarsForUnit("player")
		
	elseif event == "PLAYER_REGEN_DISABLED" or event == "PLAYER_REGEN_ENABLED" then
		self:UpdateBarsForUnit("player")
		self:UpdateBarsForUnit("target")
		self:UpdateBarsForUnit("targettarget")
	
	elseif event == "PLAYER_FOCUS_CHANGED" then
		self:UpdateBarsForUnit("focus")
		
	elseif event == "UNIT_PET" then
		self:UpdateBarsForUnit("pet")
		
	elseif event == "CHANNEL_ROSTER_UPDATE" or event == "PLAYER_ENTERING_WORLD" then
		-- List all
		for unit, _ in pairs(self.units or { }) do
			self:UpdateBarsForUnit(unit)
		end
		
	else
		self:Debug("OnTargetEvent: "..event.." General: ")
		if arg1 and arg1 ~= "player" then
			self:UpdateBarsForUnit(arg1)
		end
	end
	
end


function mHud:OnUpdate()
	
	if 1 / self.db.profile.unitFPS + self.event.lastRun > GetTime() then return end
	self.event.lastRun = GetTime()
	
	for _,v in pairs(self.onUpdateUnits) do
		if v.health ~= UnitHealth(v.unit) or v.power ~= UnitMana(v.unit) or UnitName(v.unit) ~= v.unitName then
			v.health = UnitHealth(v.unit)
			v.power = UnitMana(v.unit)
			v.unitName = UnitName(v.unit)
			self:UpdateBarsForUnit(v.unit)
			--self:Debug("OnUpdate: "..v.unit, true)
		end
	end
	
end


--[[
	Give the user a basic start. Make it easy.
]]
function mHud:SaveToDb(db, tab, pk)

	if type(tab) ~= "table" then error("mHud:SaveToDb only expect tables") end
	for k,v in pairs(tab) do
		--self:Debug(k, true)
		if type(v) ~= "table" then
			if not db[pk] then db[pk] = { } end
			db[pk][k] = v
		else
			--if not db[k] then db[k] = {} end
			self:SaveToDb(db[pk], v, k)
		end
	end
	
end


function mHud:CopyTable(to, from, pk)

	if type(from) ~= "table" or type(to) ~= "table" then error("mHud:CopyTable only expect tables (arg1 + arg2)") end
	
	if not pk then
		for k,v in pairs(from) do
			if type(v) ~= "table" then
				to[k] = v
			else
				if not to[k] then to[k] = { } end
				self:CopyTable(to, v, k)
			end
		end
	else
		for k,v in pairs(from) do
			if type(v) ~= "table" then
				to[pk][k] = v
			else
				if not to[pk][k] then to[pk][k] = { } end
				self:CopyTable(to[pk], v, k)
			end
		end
	end
	
end


function mHud:Debug(msg, force, module)

	if debug or force then
		if not module then
			module = "core"
		else
			local _, start = string.find(module, '_') -- Skip mHud_
			module = string.sub(module, start + 1)
		end
		DEFAULT_CHAT_FRAME:AddMessage("|cFFFFAA00mHud:|cFF0055DD"..module..":|cFFFFFFFF"..msg)
	end
	
end


function mHud:PrintTable(t, lvl, trav)

	if not lvl then
		self:Debug("Printing table:", true)
		lvl = 0
	end

	local lvlS = ""
	for i = 0, lvl do
		lvlS = lvlS.."---"
	end
	
	if not t then
		self:Debug(lvlS.." nil", true)
	elseif type(t) == "table" then
		for k,v in pairs(t) do
			if type(v) == "table" then
				self:Debug(lvlS.."K: "..k, true)
				if trav == nil or trav == true then
					self:PrintTable(v, lvl + 1)
				else
					self:Debug("table", true)
				end
			else
				if type(v) == "boolean" then if v == true then v = "true" else v = "false" end end
				self:Debug(lvlS.." "..k.." = "..v, true)
			end
		end
	elseif type(t) == "string" or type(t) == "number" then
		self:Debug(lvlS..t, true)
	elseif type(t) == "function" then
		self:Debug(lvlS.." Function()", true)
	else
		if type(t) == "boolean" then if t == true then t = "true" else t = "false" end end
		self:Debug(lvlS..t, true)
	end
	
end